import io
import json
import typing as t
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import fields, is_dataclass
from email.generator import Generator
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
from email.parser import FeedParser

import httpx

from async_firebase import errors
from async_firebase.errors import (
    AsyncFirebaseError,
    DeadlineExceededError,
    FcmErrorCode,
    QuotaExceededError,
    SenderIdMismatchError,
    ThirdPartyAuthError,
    UnavailableError,
    UnknownError,
    UnregisteredError,
)
from async_firebase.messages import FcmPushMulticastResponse, FcmPushResponse


def remove_null_values(dict_value: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
    """Remove Falsy values from the dictionary."""
    return {k: v for k, v in dict_value.items() if v not in [None, [], {}]}


def cleanup_firebase_message(dataclass_obj, dict_factory: t.Callable = dict) -> dict:
    """
    The instrumentation to cleanup firebase message from null values.

    Example::

        considering following dataclass

        msg = Message(
            token='qwe',
            data={},
            notification=Notification(title='push-title', body='push-body'),
            android=None,
            webpush={},
            apns=APNSConfig(
                headers={'hdr': 'qwe'},
                payload=APNSPayload(
                    aps=Aps(
                        alert=None,
                        badge=None,
                        sound='generic',
                        content_available=None,
                        category=None,
                        thread_id=None,
                        mutable_content=None,
                        custom_data={}
                    ),
                    custom_data={}
                )
            ),
            topic=None,
            condition=None
        )

        >>> dataclass_to_dict_remove_null_values(msg)
        {
            'token': 'qwe',
            'notification': {'title': 'push-title', 'body': 'push-body'},
            'apns': {
                'headers': {'hdr': 'qwe'},
                'payload': {
                    'aps': {'sound': 'generic'}
                }
            }
        }

    :param dataclass_obj: instance of dataclass. This suppose to be instance of ``messages.PushNotification`` or
        ``messages.Message``.
    :param dict_factory: if given, ``dict_factory`` will be used instead of built-in dict.
        The function applies recursively to field values that are dataclass instances.
    :return: the fields of a dataclass instance as a new dictionary mapping field names to field values.
    """
    if is_dataclass(dataclass_obj):
        result = []
        for f in fields(dataclass_obj):
            value = cleanup_firebase_message(getattr(dataclass_obj, f.name), dict_factory)
            result.append((f.name, value))
        return remove_null_values(dict_factory(result))
    elif isinstance(dataclass_obj, (list, tuple)):
        return type(dataclass_obj)(cleanup_firebase_message(v, dict_factory) for v in dataclass_obj)  # type: ignore
    elif isinstance(dataclass_obj, dict):
        return remove_null_values({k: cleanup_firebase_message(v, dict_factory) for k, v in dataclass_obj.items()})
    return deepcopy(dataclass_obj)


def serialize_mime_message(
    message: t.Union[MIMEMultipart, MIMENonMultipart],
    mangle_from: t.Optional[bool] = None,
    max_header_len: t.Optional[int] = None,
) -> str:
    """
    Serialize the MIME type message.

    :param message: MIME type message
    :param mangle_from: is a flag that, when True (the default if policy
        is not set), escapes From_ lines in the body of the message by putting
        a `>' in front of them.
    :param max_header_len: specifies the longest length for a non-continued
        header.  When a header line is longer (in characters, with tabs
        expanded to 8 spaces) than max_header_len, the header will split as
        defined in the Header class.  Set max_header_len to zero to disable
        header wrapping. The default is 78, as recommended (but not required)
        by RFC 2822.
    :return: the entire contents of the object.
    """
    fp = io.StringIO()
    gen = Generator(fp, mangle_from_=mangle_from, maxheaderlen=max_header_len)
    gen.flatten(message, unixfrom=False)
    return fp.getvalue()


FcmResponseType = t.TypeVar("FcmResponseType", FcmPushResponse, FcmPushMulticastResponse)


class FcmResponseHandler(ABC, t.Generic[FcmResponseType]):

    ERROR_CODE_TO_EXCEPTION_TYPE: t.Dict[str, t.Type[AsyncFirebaseError]] = {
        FcmErrorCode.INVALID_ARGUMENT.value: errors.InvalidArgumentError,
        FcmErrorCode.FAILED_PRECONDITION.value: errors.FailedPreconditionError,
        FcmErrorCode.OUT_OF_RANGE.value: errors.OutOfRangeError,
        FcmErrorCode.UNAUTHENTICATED.value: errors.UnauthenticatedError,
        FcmErrorCode.PERMISSION_DENIED.value: errors.PermissionDeniedError,
        FcmErrorCode.NOT_FOUND.value: errors.NotFoundError,
        FcmErrorCode.ABORTED.value: errors.AbortedError,
        FcmErrorCode.ALREADY_EXISTS.value: errors.AlreadyExistsError,
        FcmErrorCode.CONFLICT.value: errors.ConflictError,
        FcmErrorCode.RESOURCE_EXHAUSTED.value: errors.ResourceExhaustedError,
        FcmErrorCode.CANCELLED.value: errors.CancelledError,
        FcmErrorCode.DATA_LOSS.value: errors.DataLossError,
        FcmErrorCode.UNKNOWN.value: errors.UnknownError,
        FcmErrorCode.INTERNAL.value: errors.InternalError,
        FcmErrorCode.UNAVAILABLE.value: errors.UnavailableError,
        FcmErrorCode.DEADLINE_EXCEEDED.value: errors.DeadlineExceededError,
    }

    HTTP_STATUS_TO_ERROR_CODE = {
        400: FcmErrorCode.INVALID_ARGUMENT.value,
        401: FcmErrorCode.UNAUTHENTICATED.value,
        403: FcmErrorCode.PERMISSION_DENIED.value,
        404: FcmErrorCode.NOT_FOUND.value,
        409: FcmErrorCode.CONFLICT.value,
        412: FcmErrorCode.FAILED_PRECONDITION.value,
        429: FcmErrorCode.RESOURCE_EXHAUSTED.value,
        500: FcmErrorCode.INTERNAL.value,
        503: FcmErrorCode.UNAVAILABLE.value,
    }

    FCM_ERROR_TYPES = {
        "APNS_AUTH_ERROR": ThirdPartyAuthError,
        "QUOTA_EXCEEDED": QuotaExceededError,
        "SENDER_ID_MISMATCH": SenderIdMismatchError,
        "THIRD_PARTY_AUTH_ERROR": ThirdPartyAuthError,
        "UNREGISTERED": UnregisteredError,
    }

    @abstractmethod
    def handle_response(self, response: httpx.Response) -> FcmResponseType:
        pass

    @abstractmethod
    def handle_error(self, error: httpx.HTTPError) -> FcmResponseType:
        pass

    @staticmethod
    def _handle_response(response: httpx.Response) -> FcmPushResponse:
        return FcmPushResponse(fcm_response=response.json())

    def _handle_error(self, error: httpx.HTTPError) -> FcmPushResponse:
        exc = (
            (isinstance(error, httpx.HTTPStatusError) and self._handle_fcm_error(error))
            or (isinstance(error, httpx.HTTPError) and self._handle_request_error(error))
            or AsyncFirebaseError(
                code=FcmErrorCode.UNKNOWN.value,
                message="Unexpected error has happened when hitting the FCM API",
                cause=error,
            )
        )
        return FcmPushResponse(exception=exc)

    def _handle_request_error(self, error: httpx.HTTPError):
        if isinstance(error, httpx.TimeoutException):
            return DeadlineExceededError(message=f"Timed out while making an API call: {error}", cause=error)
        elif isinstance(error, httpx.ConnectError):
            return UnavailableError(message=f"Failed to establish a connection: {error}", cause=error)
        elif not hasattr(error, "response"):
            return UnknownError(message="Unknown error while making a remote service call: {error}", cause=error)

        return self._get_error_by_status_code(t.cast(httpx.HTTPStatusError, error))

    def _get_error_by_status_code(self, error: httpx.HTTPStatusError):
        error_data = self._parse_platform_error(error.response)
        code = error_data.get("status", self._http_status_to_error_code(error.response.status_code))
        err_type = self._error_code_to_exception_type(code)
        return err_type(message=error_data["message"], cause=error, http_response=error.response)  # type: ignore

    def _handle_fcm_error(self, error: httpx.HTTPStatusError):
        error_data = self._parse_platform_error(error.response)
        err_type = self._get_fcm_error_type(error_data)
        return err_type(error_data["message"], cause=error, http_response=error.response) if err_type else None

    @classmethod
    def _http_status_to_error_code(cls, http_status_code: int) -> str:
        return cls.HTTP_STATUS_TO_ERROR_CODE.get(http_status_code, FcmErrorCode.UNKNOWN.value)

    @classmethod
    def _error_code_to_exception_type(cls, error_code: str) -> t.Type[AsyncFirebaseError]:
        return cls.ERROR_CODE_TO_EXCEPTION_TYPE.get(error_code, errors.UnknownError)

    @classmethod
    def _get_fcm_error_type(cls, error_data: dict):
        if not error_data:
            return None

        fcm_code = None
        for detail in error_data.get("details", []):
            if detail.get("@type") == "type.googleapis.com/google.firebase.fcm.v1.FcmError":
                fcm_code = detail.get("errorCode")
                break

        if not fcm_code:
            return None

        return cls.FCM_ERROR_TYPES.get(fcm_code)

    @staticmethod
    def _parse_platform_error(response: httpx.Response):
        """Extract the code and mesage from GCP API Error HTTP response."""
        data: dict = {}
        try:
            data = response.json()
        except ValueError:
            pass

        error_data = data.get("error", {})
        if not error_data.get("message"):
            error_data[
                "message"
            ] = f"Unexpected HTTP response with status: {response.status_code}; body: {response.content!r}"
        return error_data


class FcmPushResponseHandler(FcmResponseHandler[FcmPushResponse]):
    def handle_error(self, error: httpx.HTTPError) -> FcmPushResponse:
        return self._handle_error(error)

    def handle_response(self, response: httpx.Response) -> FcmPushResponse:
        return self._handle_response(response)


class FcmPushMulticastResponseHandler(FcmResponseHandler[FcmPushMulticastResponse]):
    def handle_error(self, error: httpx.HTTPError):
        fcm_response = self._handle_error(error)
        return FcmPushMulticastResponse(responses=[fcm_response])

    def handle_response(self, response: httpx.Response):
        fcm_push_responses = []
        responses = self._deserialize_batch_response(response)
        for single_resp in responses:
            if single_resp.status_code >= 300:
                exc = httpx.HTTPStatusError("FCM Error", response=single_resp, request=response.request)
                fcm_push_responses.append(self._handle_error(exc))
            else:
                fcm_push_responses.append(self._handle_response(single_resp))

        return FcmPushMulticastResponse(responses=fcm_push_responses)

    @staticmethod
    def _deserialize_batch_response(response: httpx.Response) -> t.List[httpx.Response]:
        """Convert batch response into list of `httpx.Response` responses for each multipart.

        :param response: string, headers and body as a string.
        :return: list of `httpx.Response` responses.
        """
        # Prepend with a content-type header so FeedParser can handle it.
        header = f"content-type: {response.headers['content-type']}\r\n\r\n"
        # PY3's FeedParser only accepts unicode. So we should decode content here, and encode each payload again.
        content = response.content.decode()
        for_parser = f"{header}{content}"

        parser = FeedParser()
        parser.feed(for_parser)
        mime_response = parser.close()

        if not mime_response.is_multipart():
            raise ValueError("Response not in multipart/mixed format.")

        responses = []
        for part in mime_response.get_payload():
            request_id = part["Content-ID"].split("-", 1)[-1]
            status_line, payload = part.get_payload().split("\n", 1)
            _, status_code, _ = status_line.split(" ", 2)
            status_code = int(status_code)

            # Parse the rest of the response
            parser = FeedParser()
            parser.feed(payload)
            msg = parser.close()
            msg["status_code"] = status_code

            # Create httpx.Response from the parsed headers.
            resp = httpx.Response(
                status_code=status_code,
                headers=httpx.Headers({"Content-Type": msg.get_content_type(), "X-Request-ID": request_id}),
                content=msg.get_payload(),
                json=json.loads(msg.get_payload()),
            )
            responses.append(resp)

        return responses



"""
The module houses client to communicate with FCM - Firebase Cloud Messaging (Android, iOS and Web).

Documentation for google-auth package https://google-auth.readthedocs.io/en/latest/user-guide.html that is used
to authorize request which is being made to Firebase.
"""
import json
import logging
import typing as t
import uuid
from dataclasses import replace
from datetime import datetime, timedelta
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
from pathlib import PurePath
from urllib.parse import urlencode, urljoin

import httpx
from google.oauth2 import service_account  # type: ignore

import pkg_resources  # type: ignore
from async_firebase.encoders import aps_encoder
from async_firebase.messages import (
    AndroidConfig,
    AndroidNotification,
    APNSConfig,
    APNSPayload,
    Aps,
    ApsAlert,
    FcmPushMulticastResponse,
    FcmPushResponse,
    Message,
    Notification,
    PushNotification,
    WebpushConfig,
    WebpushFCMOptions,
    WebpushNotification,
    WebpushNotificationAction,
)
from async_firebase.utils import (
    FcmPushMulticastResponseHandler,
    FcmPushResponseHandler,
    cleanup_firebase_message,
    serialize_mime_message,
)

DEFAULT_TTL = 604800
MULTICAST_MESSAGE_MAX_DEVICE_TOKENS = 500


class AsyncFirebaseClient:
    """Async wrapper for Firebase Cloud Messaging.

    The AsyncFirebaseClient relies on Service Account to enable us making a request. To get more about Service Account
    please refer to https://firebase.google.com/support/guides/service-accounts
    """

    BASE_URL: str = "https://fcm.googleapis.com"
    TOKEN_URL: str = "https://oauth2.googleapis.com/token"
    FCM_ENDPOINT: str = "/v1/projects/{project_id}/messages:send"
    FCM_BATCH_ENDPOINT: str = "/batch"
    # A list of accessible OAuth 2.0 scopes can be found https://developers.google.com/identity/protocols/oauth2/scopes.
    SCOPES: t.List[str] = [
        "https://www.googleapis.com/auth/cloud-platform",
    ]

    def __init__(
        self,
        credentials: t.Optional[service_account.Credentials] = None,
        scopes: t.Optional[t.List[str]] = None,
    ) -> None:
        """
        :param credentials: instance of ``google.oauth2.service_account.Credentials``.
            Usually, you'll create these credentials with one of the helper constructors. To create credentials using a
            Google service account private key JSON file::

                self.creds_from_service_account_file('service-account.json')

            Or if you already have the service account file loaded::

                service_account_info = json.load(open('service_account.json'))
                self.creds_from_service_account_info(service_account_info)

        :param scopes: user-defined scopes to request during the authorization grant.
        """
        self._credentials: service_account.Credentials = credentials
        self.scopes: t.List[str] = scopes or self.SCOPES

    @staticmethod
    def assemble_push_notification(
        *,
        apns_config: t.Optional[APNSConfig],
        message: Message,
        dry_run: bool,
    ) -> t.Dict[str, t.Any]:
        """
        Assemble ``messages.PushNotification`` object properly.

        :param apns_config: instance of ``messages.APNSConfig``
        :param dry_run: A boolean indicating whether to run the operation in dry run mode
        :param message: an instance of ``messages.Message``
        :return: dictionary with push notification data ready to send
        """
        has_apns_config = True if apns_config and apns_config.payload else False
        if has_apns_config:
            # avoid mutation of active message
            message.apns = replace(message.apns)
            message.apns.payload = aps_encoder(apns_config.payload.aps)  # type: ignore

        push_notification: t.Dict[str, t.Any] = cleanup_firebase_message(
            PushNotification(message=message, validate_only=dry_run)
        )
        if len(push_notification["message"]) == 1:
            logging.warning("No data has been provided to construct push notification payload")
            raise ValueError("``messages.PushNotification`` cannot be assembled as data has not been provided")
        return push_notification

    def creds_from_service_account_info(self, service_account_info: t.Dict[str, str]) -> None:
        """
        Creates a Credentials instance from parsed service account info.

        :param service_account_info: the service account info in Google format.
        """
        self._credentials = service_account.Credentials.from_service_account_info(
            info=service_account_info, scopes=self.scopes
        )

    def creds_from_service_account_file(self, service_account_filename: t.Union[str, PurePath]) -> None:
        """
        Creates a Credentials instance from a service account json file.

        :param service_account_filename: the path to the service account json file.
        """
        if isinstance(service_account_filename, PurePath):
            service_account_filename = str(service_account_filename)

        logging.debug("Creating credentials from file: %s", service_account_filename)
        self._credentials = service_account.Credentials.from_service_account_file(
            filename=service_account_filename, scopes=self.scopes
        )

    async def _get_access_token(self) -> str:
        """Get OAuth 2 access token."""
        if self._credentials.valid:
            return self._credentials.token

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = urlencode(
            {
                "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
                "assertion": self._credentials._make_authorization_grant_assertion(),
            }
        ).encode("utf-8")

        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.post(self.TOKEN_URL, data=data, headers=headers)
            response_data = response.json()

        self._credentials.expiry = datetime.utcnow() + timedelta(seconds=response_data["expires_in"])
        self._credentials.token = response_data["access_token"]
        return self._credentials.token

    @staticmethod
    def build_android_config(  # pylint: disable=too-many-locals
        *,
        priority: str,
        ttl: t.Union[int, timedelta] = DEFAULT_TTL,
        collapse_key: t.Optional[str] = None,
        restricted_package_name: t.Optional[str] = None,
        data: t.Optional[t.Dict[str, t.Any]] = None,
        title: t.Optional[str] = None,
        body: t.Optional[str] = None,
        icon: t.Optional[str] = None,
        color: t.Optional[str] = None,
        sound: t.Optional[str] = None,
        tag: t.Optional[str] = None,
        click_action: t.Optional[str] = None,
        body_loc_key: t.Optional[str] = None,
        body_loc_args: t.Optional[t.List[str]] = None,
        title_loc_key: t.Optional[str] = None,
        title_loc_args: t.Optional[t.List[str]] = None,
        channel_id: t.Optional[str] = None,
        notification_count: t.Optional[int] = None,
    ) -> AndroidConfig:
        """
        Constructs AndroidConfig that will be used to customize the messages that are sent to Android device.

        :param priority: sets the priority of the message. Valid values are "normal" and "high."
        :param ttl: this parameter specifies how long (in seconds) the message should be kept in Firebase storage if the
            device is offline. The maximum time to live supported is 4 weeks, and the default value is 4 weeks.
        :param collapse_key: this parameter identifies a group of messages that can be collapsed, so that only the last
            message gets sent when delivery can be resumed.
        :param restricted_package_name: The package name of the application where the registration tokens must match in
            order to receive the message (optional).
        :param data: A dictionary of data fields (optional). All keys and values in the dictionary must be strings.
        :param title: Title of the notification (optional).
        :param body: Body of the notification (optional).
        :param icon: Icon of the notification (optional).
        :param color: Color of the notification icon expressed in ``#rrggbb`` form (optional).
        :param sound: Sound to be played when the device receives the notification (optional). This is usually the file
            name of the sound resource.
        :param tag: Tag of the notification (optional). This is an identifier used to replace existing notifications in
            the notification drawer. If not specified, each request creates a new notification.
        :param click_action: The action associated with a user click on the notification (optional). If specified, an
            activity with a matching intent filter is launched when a user clicks on the notification.
        :param body_loc_key: Key of the body string in the app's string resources to use to localize the
            body text (optional).
        :param body_loc_args: A list of resource keys that will be used in place of the format specifiers
            in ``body_loc_key`` (optional).
        :param title_loc_key: Key of the title string in the app's string resources to use to localize the
            title text (optional).
        :param title_loc_args: A list of resource keys that will be used in place of the format specifiers
            in ``title_loc_key`` (optional).
        :param channel_id: Notification channel id, used by android to allow user to configure notification display
            rules on per-channel basis (optional).
        :param notification_count: The number of items in notification. May be displayed as a badge count for launchers
            that support badging. If zero or unspecified, systems that support badging use the default, which is to
            increment a number displayed on the long-press menu each time a new notification arrives (optional).
        :return: an instance of ``messages.AndroidConfig`` to be included in the resulting payload.
        """
        android_config = AndroidConfig(
            collapse_key=collapse_key,
            priority=priority,
            ttl=f"{int(ttl.total_seconds()) if isinstance(ttl, timedelta) else ttl}s",
            restricted_package_name=restricted_package_name,
            data={str(key): "null" if value is None else str(value) for key, value in data.items()} if data else {},
            notification=AndroidNotification(
                title=title,
                body=body,
                icon=icon,
                color=color,
                sound=sound,
                tag=tag,
                click_action=click_action,
                body_loc_key=body_loc_key,
                body_loc_args=body_loc_args or [],
                title_loc_key=title_loc_key,
                title_loc_args=title_loc_args or [],
                channel_id=channel_id,
                notification_count=notification_count,
            ),
        )

        return android_config

    @staticmethod
    def build_apns_config(  # pylint: disable=too-many-locals
        *,
        priority: str,
        ttl: int = DEFAULT_TTL,
        apns_topic: t.Optional[str] = None,
        collapse_key: t.Optional[str] = None,
        title: t.Optional[str] = None,
        alert: t.Optional[str] = None,
        badge: t.Optional[int] = None,
        sound: t.Optional[str] = None,
        content_available: bool = False,
        category: t.Optional[str] = None,
        thread_id: t.Optional[str] = None,
        mutable_content: bool = True,
        custom_data: t.Optional[t.Dict[str, t.Any]] = None,
        loc_key: t.Optional[str] = None,
        loc_args: t.Optional[t.List[str]] = None,
        title_loc_key: t.Optional[str] = None,
        title_loc_args: t.Optional[t.List[str]] = None,
        action_loc_key: t.Optional[str] = None,
        launch_image: t.Optional[str] = None,
    ) -> APNSConfig:
        """
        Constructs APNSConfig that will be used to customize the messages that are sent to iOS device.

        :param priority: sets the priority of the message. On iOS, these correspond to APNs priorities 5 and 10.
        :param ttl: this parameter specifies how long (in seconds) the message should be kept in Firebase storage if the
            device is offline. The maximum time to live supported is 4 weeks, and the default value is 4 weeks.
        :param apns_topic: the topic for the notification. In general, the topic is your app’s bundle ID/app ID.
            It can have a suffix based on the type of push notification.
        :param collapse_key: this parameter identifies a group of messages that can be collapsed, so that only the last
            message gets sent when delivery can be resumed.
        :param title: title of the alert (optional). If specified, overrides the title set via ``messages.Notification``
        :param alert: a string or a ``messages.ApsAlert`` instance (optional).
        :param badge: the value of the badge on the home screen app icon. If not specified, the badge is not changed.
            If set to 0, the badge is removed.
        :param sound: name of the sound file to be played with the message (optional).
        :param content_available: A boolean indicating whether to configure a background update notification (optional).
        :param category: string identifier representing the message type (optional).
        :param thread_id: an app-specific string identifier for grouping messages (optional).
        :param mutable_content: A boolean indicating whether to support mutating notifications at the client using app
            extensions (optional).
        :param custom_data: A dict of custom key-value pairs to be included in the Aps dictionary (optional).
        :param loc_key: key of the body string in the app's string resources to use to localize the body text
            (optional).
        :param loc_args: a list of resource keys that will be used in place of the format specifiers in ``loc_key``
            (optional).
        :param title_loc_key: key of the title string in the app's string resources to use to localize the title text
            (optional).
        :param title_loc_args: a list of resource keys that will be used in place of the format specifiers in
            ``title_loc_key`` (optional).
        :param action_loc_key: key of the text in the app's string resources to use to localize the action button text
            (optional).
        :param launch_image: image for the notification action (optional).
        :return: an instance of ``messages.APNSConfig`` to included in the resulting payload.
        """

        apns_headers = {
            "apns-expiration": str(int(datetime.utcnow().timestamp()) + ttl),
            "apns-priority": str(10 if priority == "high" else 5),
        }
        if apns_topic:
            apns_headers["apns-topic"] = apns_topic
        if collapse_key:
            apns_headers["apns-collapse-id"] = str(collapse_key)

        apns_config = APNSConfig(
            headers=apns_headers,
            payload=APNSPayload(
                aps=Aps(
                    alert=ApsAlert(
                        title=title,
                        body=alert,
                        loc_key=loc_key,
                        loc_args=loc_args or [],
                        title_loc_key=title_loc_key,
                        title_loc_args=title_loc_args or [],
                        action_loc_key=action_loc_key,
                        launch_image=launch_image,
                    ),
                    badge=badge,
                    sound="default" if alert and sound is None else sound,
                    category=category,
                    thread_id=thread_id,
                    mutable_content=mutable_content,
                    custom_data=custom_data or {},
                    content_available=True if content_available else None,
                ),
            ),
        )

        return apns_config

    @staticmethod
    def build_webpush_config(  # pylint: disable=too-many-locals
        *,
        data: t.Dict[str, str],
        headers: t.Optional[t.Dict[str, str]] = None,
        title: t.Optional[str] = None,
        body: t.Optional[str] = None,
        icon: t.Optional[str] = None,
        actions: t.Optional[t.List[WebpushNotificationAction]] = None,
        badge: t.Optional[str] = None,
        direction: t.Optional[str] = "auto",
        image: t.Optional[str] = None,
        language: t.Optional[str] = None,
        renotify: t.Optional[bool] = False,
        require_interaction: t.Optional[bool] = None,
        silent: t.Optional[bool] = False,
        tag: t.Optional[str] = None,
        timestamp_millis: t.Optional[int] = None,
        vibrate: t.Optional[str] = None,
        custom_data: t.Optional[t.Dict[str, str]] = None,
        link: t.Optional[str] = None,
    ) -> WebpushConfig:
        """
        Constructs WebpushConfig that will be used to customize the messages that are sent user agents.

        :param data: A dictionary of data fields (optional). All keys and values in the dictionary must be strings.
        :param headers: a dictionary of headers (optional).
        :param title: title of the notification (optional).
        :param body: body of the notification (optional).
        :param icon: icon URL of the notification (optional).
        :param actions: a list of ``messages.WebpushNotificationAction`` instances (optional).
        :param badge: URL of the image used to represent the notification when there is not enough space to display the
            notification itself (optional).
        :param data: any arbitrary JSON data that should be associated with the notification (optional).
        :param direction: the direction in which to display the notification (optional). Must be either 'auto', 'ltr'
            or 'rtl'.
        :param image: the URL of an image to be displayed in the notification (optional).
        :param language: notification language (optional).
        :param renotify: a boolean indicating whether the user should be notified after a new notification replaces
            an old one (optional).
        :param require_interaction: a boolean indicating whether a notification should remain active until the user
            clicks or dismisses it, rather than closing automatically (optional).
        :param silent: ``True`` to indicate that the notification should be silent (optional).
        :param tag: an identifying tag on the notification (optional).
        :param timestamp_millis: a timestamp value in milliseconds on the notification (optional).
        :param vibrate: a vibration pattern for the device's vibration hardware to emit when the notification
            fires (optional). The pattern is specified as an integer array.
        :param custom_data: a dict of custom key-value pairs to be included in the notification (optional)
        :param link: The link to open when the user clicks on the notification. Must be an HTTPS URL (optional).
        :return: an instance of ``messages.WebpushConfig`` to included in the resulting payload.
        """
        return WebpushConfig(
            data=data,
            headers=headers or {},
            notification=WebpushNotification(
                title=title,
                body=body,
                icon=icon,
                actions=actions or [],
                badge=badge,
                direction=direction,
                image=image,
                language=language,
                renotify=renotify,
                require_interaction=require_interaction,
                silent=silent,
                tag=tag,
                timestamp_millis=timestamp_millis,
                vibrate=vibrate,
                custom_data=custom_data or {},
            ),
            fcm_options=WebpushFCMOptions(link=link) if link else None,
        )

    @staticmethod
    def _get_request_id():
        """Generate unique request ID."""
        return str(uuid.uuid4())

    @staticmethod
    def _serialize_batch_request(request: httpx.Request) -> str:
        """
        Convert an HttpRequest object into a string.

        :param request: `httpx.Request`, the request to serialize.
        :return: a string in application/http format.
        """
        status_line = f"{request.method} {request.url.path} HTTP/1.1\n"
        major, minor = request.headers.get("content-type", "application/json").split("/")
        msg = MIMENonMultipart(major, minor)
        headers = request.headers.copy()

        # MIMENonMultipart adds its own Content-Type header.
        if "content-type" in headers:
            del headers["content-type"]

        for key, value in headers.items():
            msg[key] = value
        msg.set_unixfrom(None)  # type: ignore

        if request.content is not None:
            msg.set_payload(request.content)
            msg["content-length"] = str(len(request.content))

        body = serialize_mime_message(msg, max_header_len=0)
        return f"{status_line}{body}"

    async def _prepare_headers(self):
        """Prepare HTTP headers that will be used to request Firebase Cloud Messaging."""
        logging.debug("Preparing HTTP headers for all the subsequent requests")
        access_token: str = await self._get_access_token()
        return {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json; UTF-8",
            "X-Request-Id": self._get_request_id(),
            "X-GOOG-API-FORMAT-VERSION": "2",
            "X-FIREBASE-CLIENT": "async-firebase/{0}".format(pkg_resources.get_distribution("async-firebase").version),
        }

    async def push(  # pylint: disable=too-many-locals
        self,
        device_token: str,
        *,
        android: t.Optional[AndroidConfig] = None,
        apns: t.Optional[APNSConfig] = None,
        condition: t.Optional[str] = None,
        data: t.Optional[t.Dict[str, str]] = None,
        notification: t.Optional[Notification] = None,
        topic: t.Optional[str] = None,
        webpush: t.Optional[WebpushConfig] = None,
        dry_run: bool = False,
    ) -> FcmPushResponse:
        """
        Send push notification.

        :param device_token: device token allows to send targeted notifications to a particular instance of app.
        :param android: as instance of ``messages.AndroidConfig`` that contains Android-specific options.
        :param apns: as instance of ``messages.APNSConfig`` that contains iOS-specific options.
        :param condition: the Firebase condition to which the message should be sent.
        :param data: a dictionary of data fields. All keys and values in the dictionary must be strings.
        :param notification: an instance of ``messages.Notification`` that contains a notification that can be included
            in a resulting message.
        :param topic: name of the Firebase topic to which the message should be sent.
        :param webpush: an instance of ``messages.WebpushConfig``.
        :param dry_run: indicating whether to run the operation in dry run mode (optional). Flag for testing the request
            without actually delivering the message. Default to ``False``.

        :raises:

            ValueError is ``messages.PushNotification`` payload cannot be assembled

        :return: response from Firebase. Example of response:

            success::

                {
                    'name': 'projects/mobile-app/messages/0:1612788010922733%7606eb247606eb24'
                }

            failure::

                {
                    'error': {
                        'code': 400,
                        'details': [
                            {
                                '@type': 'type.googleapis.com/google.rpc.BadRequest',
                                'fieldViolations': [
                                    {
                                        'description': 'Value type for APS key [badge] is a number.',
                                        'field': 'message.apns.payload.aps.badge'
                                    }
                                ]
                            },
                            {
                                '@type': 'type.googleapis.com/google.firebase.fcm.v1.FcmError',
                                'errorCode': 'INVALID_ARGUMENT'
                            }
                        ],
                        'message': 'Value type for APS key [badge] is a number.',
                        'status': 'INVALID_ARGUMENT'
                    }
                }
        """
        message = Message(
            token=device_token,
            data=data or {},
            notification=notification,
            android=android,
            webpush=webpush,
            apns=apns,
            topic=topic,
            condition=condition,
        )

        push_notification = self.assemble_push_notification(apns_config=apns, dry_run=dry_run, message=message)

        response = await self._send_request(
            uri=self.FCM_ENDPOINT.format(project_id=self._credentials.project_id),  # type: ignore
            json_payload=push_notification,
            headers=await self._prepare_headers(),
            response_handler=FcmPushResponseHandler(),
        )
        if not isinstance(response, FcmPushResponse):
            raise ValueError("Wrong return type, perhaps because of a response handler misuse.")
        return response

    async def push_multicast(
        self,
        device_tokens: t.Union[t.List[str], t.Tuple[str]],
        *,
        android: t.Optional[AndroidConfig] = None,
        apns: t.Optional[APNSConfig] = None,
        data: t.Optional[t.Dict[str, str]] = None,
        notification: t.Optional[Notification] = None,
        webpush: t.Optional[WebpushConfig] = None,
        dry_run: bool = False,
    ) -> FcmPushMulticastResponse:
        """
        Send Multicast push notification.

        :param device_tokens: the list of device tokens to send targeted notifications to a set of instances of app.
            May contain up to 500 device tokens.
        :param android: as instance of ``messages.AndroidConfig`` that contains Android-specific options.
        :param apns: as instance of ``messages.APNSConfig`` that contains iOS-specific options.
        :param data: a dictionary of data fields. All keys and values in the dictionary must be strings.
        :param notification: an instance of ``messages.Notification`` that contains a notification that can be included
            in a resulting message.
        :param webpush: an instance of ``messages.WebpushConfig``.
        :param dry_run: indicating whether to run the operation in dry run mode (optional). Flag for testing the request
            without actually delivering the message. Default to ``False``.
        """

        if len(device_tokens) > MULTICAST_MESSAGE_MAX_DEVICE_TOKENS:
            raise ValueError(
                f"A single ``messages.MulticastMessage`` may contain up to {MULTICAST_MESSAGE_MAX_DEVICE_TOKENS} "
                "device tokens."
            )

        multipart_message = MIMEMultipart("mixed")
        # Message should not write out it's own headers.
        setattr(multipart_message, "_write_headers", lambda self: None)

        for device_token in device_tokens:
            msg = MIMENonMultipart("application", "http")
            msg["Content-Transfer-Encoding"] = "binary"
            msg["Content-ID"] = self._get_request_id()
            push_notification = self.assemble_push_notification(
                apns_config=apns,
                dry_run=dry_run,
                message=Message(
                    token=device_token,
                    data=data or {},
                    notification=notification,
                    android=android,
                    webpush=webpush,
                    apns=apns,
                ),
            )
            body = self._serialize_batch_request(
                httpx.Request(
                    method="POST",
                    url=urljoin(
                        self.BASE_URL, self.FCM_ENDPOINT.format(project_id=self._credentials.project_id)  # type: ignore
                    ),
                    headers=await self._prepare_headers(),
                    content=json.dumps(push_notification),
                )
            )
            msg.set_payload(body)
            multipart_message.attach(msg)

        # encode the body: note that we can't use `as_string`, because it plays games with `From ` lines.
        body = serialize_mime_message(multipart_message, mangle_from=False)

        batch_response = await self._send_request(
            uri=self.FCM_BATCH_ENDPOINT,
            content=body,
            headers={"Content-Type": f"multipart/mixed; boundary={multipart_message.get_boundary()}"},
            response_handler=FcmPushMulticastResponseHandler(),
        )
        if not isinstance(batch_response, FcmPushMulticastResponse):
            raise ValueError("Wrong return type, perhaps because of a response handler misuse.")
        return batch_response

    async def _send_request(
        self,
        uri: str,
        response_handler: t.Union[FcmPushResponseHandler, FcmPushMulticastResponseHandler],
        json_payload: t.Optional[t.Dict[str, t.Any]] = None,
        headers: t.Optional[t.Dict[str, str]] = None,
        content: t.Union[str, bytes, t.Iterable[bytes], t.AsyncIterable[bytes], None] = None,
    ) -> t.Union[FcmPushResponse, FcmPushMulticastResponse]:
        """
        Sends an HTTP call using the ``httpx`` library.

        :param uri: URI to be requested.
        :param json_payload: request JSON payload
        :param headers: request headers.
        :param content: request content
        :return: HTTP response
        """
        async with httpx.AsyncClient(base_url=self.BASE_URL) as client:
            logging.debug(
                "Requesting POST %s, payload: %s, content: %s, headers: %s",
                urljoin(self.BASE_URL, self.FCM_ENDPOINT.format(project_id=self._credentials.project_id)),
                json_payload,
                content,
                headers,
            )
            try:
                raw_fcm_response: httpx.Response = await client.post(
                    uri,
                    json=json_payload,
                    headers=headers or await self._prepare_headers(),
                    content=content,
                )
                raw_fcm_response.raise_for_status()
            except httpx.HTTPError as exc:
                response = response_handler.handle_error(exc)
            else:
                logging.debug(
                    "Response Code: %s, Time spent to make a request: %s",
                    raw_fcm_response.status_code,
                    raw_fcm_response.elapsed,
                )
                response = response_handler.handle_response(raw_fcm_response)

        return response



"""The module houses the message structures that can be used to construct a push notification payload.

"""

import typing as t
from dataclasses import dataclass, field

from async_firebase.errors import AsyncFirebaseError


@dataclass
class AndroidNotification:
    """
    Android-specific notification parameters.

    Attributes:
    title: title of the notification (optional). If specified, overrides the title set via ``messages.Notification``.
    body: body of the notification (optional). If specified, overrides the body set via ``messages.Notification``.
    icon: icon of the notification (optional).
    color: color of the notification icon expressed in ``#rrggbb`` form (optional).
    sound: sound to be played when the device receives the notification (optional). This is usually the file name of
        the sound resource.
    tag: tag of the notification (optional). This is an identifier used to replace existing notifications in the
        notification drawer. If not specified, each request creates a new notification.
    click_action: the action associated with a user click on the notification (optional). If specified, an activity with
        a matching intent filter is launched when a user clicks on the notification.
    body_loc_key: key of the body string in the app's string resources to use to localize the body text (optional).
    body_loc_args: a list of resource keys that will be used in place of the format specifiers in ``body_loc_key``
        (optional).
    title_loc_key: key of the title string in the app's string resources to use to localize the title text (optional).
    title_loc_args: a list of resource keys that will be used in place of the format specifiers in ``title_loc_key``
        (optional).
    channel_id: The notification's channel id (new in Android O). The app must create a channel with this channel ID
        before any notification with this channel ID is received. If you don't send this channel ID in the request,
        or if the channel ID provided has not yet been created by the app, FCM uses the channel ID specified in the
        app manifest.
    notification_count: the number of items this notification represents (optional). If zero or unspecified, systems
        that support badging use the default, which is to increment a number displayed on the long-press menu each time
        a new notification arrives.
    """

    title: t.Optional[str] = None
    body: t.Optional[str] = None
    icon: t.Optional[str] = None
    color: t.Optional[str] = None
    sound: t.Optional[str] = None
    tag: t.Optional[str] = None
    click_action: t.Optional[str] = None
    body_loc_key: t.Optional[str] = None
    body_loc_args: t.List[str] = field(default_factory=list)
    title_loc_key: t.Optional[str] = None
    title_loc_args: t.List[str] = field(default_factory=list)
    channel_id: t.Optional[str] = None
    notification_count: t.Optional[int] = None


@dataclass
class AndroidConfig:
    """
    Android-specific options that can be included in a message.

    Attributes:
    collapse_key: collapse key string for the message (optional). This is an identifier for a group of messages that
        can be collapsed, so that only the last message is sent when delivery can be resumed. A maximum of 4 different
        collapse keys may be active at a given time.
    priority: priority of the message (optional). Must be one of ``high`` or ``normal``.
    ttl: the time-to-live duration of the message (optional) represent as string. For example: 7200s
    restricted_package_name: the package name of the application where the registration tokens must match in order to
        receive the message (optional).
    data: a dictionary of data fields (optional). All keys and values in the dictionary must be strings.
    notification: a ``messages.AndroidNotification`` to be included in the message (optional).
    """

    collapse_key: t.Optional[str] = None
    priority: t.Optional[str] = None
    ttl: t.Optional[str] = None
    restricted_package_name: t.Optional[str] = None
    data: t.Dict[str, str] = field(default_factory=dict)
    notification: t.Optional[AndroidNotification] = field(default=None)


@dataclass
class ApsAlert:
    """
    An alert that can be included in ``message.Aps``.

    Attributes:
    title: title of the alert (optional). If specified, overrides the title set via ``messages.Notification``.
    body: body of the alert (optional). If specified, overrides the body set via ``messages.Notification``.
    loc_key: key of the body string in the app's string resources to use to localize the body text (optional).
    loc_args: a list of resource keys that will be used in place of the format specifiers in ``loc_key`` (optional).
    title_loc_key: key of the title string in the app's string resources to use to localize the title text (optional).
    title_loc_args: a list of resource keys that will be used in place of the format specifiers in ``title_loc_key``
        (optional).
    action_loc_key: key of the text in the app's string resources to use to localize the action button text (optional).
    launch_image: image for the notification action (optional).
    """

    title: t.Optional[str] = None
    body: t.Optional[str] = None
    loc_key: t.Optional[str] = None
    loc_args: t.List[str] = field(default_factory=list)
    title_loc_key: t.Optional[str] = None
    title_loc_args: t.List[str] = field(default_factory=list)
    action_loc_key: t.Optional[str] = None
    launch_image: t.Optional[str] = None


@dataclass
class Aps:
    """
    Aps dictionary to be included in an APNS payload.

    Attributes:
    alert: a string or a ``messages.ApsAlert`` instance (optional).
    badge: a number representing the badge to be displayed with the message (optional).
    sound: name of the sound file to be played with the message (optional).
    content_available: a boolean indicating whether to configure a background update notification (optional).
    category: string identifier representing the message type (optional).
    thread_id: an app-specific string identifier for grouping messages (optional).
    mutable_content: a boolean indicating whether to support mutating notifications at the client using app extensions
        (optional).
    custom_data: a dictionary of custom key-value pairs to be included in the Aps dictionary (optional). These
        attributes will be then re-assembled according to the format allowed by APNS. Below you may find details:

        In addition to the Apple-defined keys, custom keys may be added to payload to deliver small amounts of data
        to app, notification service app extension, or notification content app extension. Custom keys must
        have values with primitive types, such as dictionary, array, string, number, or Boolean. Custom keys are
        available in the ``userInfo`` dictionary of the ``UNNotificationContent`` object delivered to app.

        In a nutshell custom keys should be incorporated on the same level as ``apns`` attribute.
    """

    alert: t.Union[str, ApsAlert, None] = None
    badge: t.Optional[int] = None
    sound: t.Optional[str] = None
    content_available: t.Optional[bool] = None
    category: t.Optional[str] = None
    thread_id: t.Optional[str] = None
    mutable_content: t.Optional[bool] = None
    custom_data: t.Dict[str, str] = field(default_factory=dict)


@dataclass
class APNSPayload:
    """
    Payload of an APNS message.

    Attributes:
    aps: a ``messages.Aps`` instance to be included in the payload.
    """

    aps: t.Optional[Aps] = field(default=None)


@dataclass
class APNSConfig:
    """
    APNS-specific options that can be included in a message.

    Refer to APNS Documentation: https://developer.apple.com/library/content/documentation\
        /NetworkingInternet/Conceptual/RemoteNotificationsPG/CommunicatingwithAPNs.html for more information.

    Attributes:
    headers: a dictionary of headers (optional).
    payload: a ``messages.APNSPayload`` to be included in the message (optional).
    """

    headers: t.Dict[str, str] = field(default_factory=dict)
    payload: t.Optional[APNSPayload] = field(default=None)


@dataclass
class Notification:
    """
    A notification that can be included in a message.

    Attributes:
    title: title of the notification (optional).
    body: body of the notification (optional).
    image: image url of the notification (optional)
    """

    title: t.Optional[str] = None
    body: t.Optional[str] = None
    image: t.Optional[str] = None


@dataclass
class WebpushFCMOptions:
    """
    Options for features provided by the FCM SDK for Web.

    Arguments:
    link: The link to open when the user clicks on the notification. Must be an HTTPS URL (optional).
    """

    link: str


@dataclass
class WebpushNotificationAction:
    """
    An action available to the users when the notification is presented.

    Arguments:
    action: Action string.
    title: Title string.
    icon: Icon URL for the action (optional).
    """

    action: t.Optional[str]
    title: t.Optional[str]
    icon: t.Optional[str] = None


@dataclass
class WebpushNotification:
    """
    Webpush-specific notification parameters.

    Arguments:
    title: title of the notification (optional). If specified, overrides the title set via ``messages.Notification``.
    body: body of the notification (optional). If specified, overrides the body set via ``messages.Notification``.
    icon: icon URL of the notification (optional).
    actions: a list of ``messages.WebpushNotificationAction`` instances (optional).
    badge: URL of the image used to represent the notification when there is not enough space to display the
        notification itself (optional).
    data: any arbitrary JSON data that should be associated with the notification (optional).
    direction: the direction in which to display the notification (optional). Must be either 'auto', 'ltr' or 'rtl'.
    image: the URL of an image to be displayed in the notification (optional).
    language: notification language (optional).
    renotify: a boolean indicating whether the user should be notified after a new notification replaces
        an old one (optional).
    require_interaction: a boolean indicating whether a notification should remain active until the user clicks or
        dismisses it, rather than closing automatically (optional).
    silent: ``True`` to indicate that the notification should be silent (optional).
    tag: an identifying tag on the notification (optional).
    timestamp_millis: a timestamp value in milliseconds on the notification (optional).
    vibrate: a vibration pattern for the device's vibration hardware to emit when the notification fires (optional).
        The pattern is specified as an integer array.
    custom_data: a dict of custom key-value pairs to be included in the notification (optional)
    """

    title: t.Optional[str] = None
    body: t.Optional[str] = None
    icon: t.Optional[str] = None
    actions: t.List[WebpushNotificationAction] = field(default_factory=list)
    badge: t.Optional[str] = None
    data: t.Dict[str, str] = field(default_factory=dict)
    direction: t.Optional[str] = None
    image: t.Optional[str] = None
    language: t.Optional[str] = None
    renotify: t.Optional[bool] = None
    require_interaction: t.Optional[bool] = None
    silent: t.Optional[bool] = None
    tag: t.Optional[str] = None
    timestamp_millis: t.Optional[int] = None
    vibrate: t.Optional[str] = None
    custom_data: t.Dict[str, str] = field(default_factory=dict)


@dataclass
class WebpushConfig:
    """
    Webpush-specific options that can be included in a message.

    Attributes:
    headers: a dictionary of headers (optional). Refer to
        [Webpush protocol](https://tools.ietf.org/html/rfc8030#section-5) for supported headers.
    data: A dictionary of data fields (optional). All keys and values in the dictionary must be
        strings. When specified, overrides any data fields set via ``Message.data``.
    notification: a ``messages.WebpushNotification`` to be included in the message (optional).
    fcm_options: a ``messages.WebpushFCMOptions`` instance to be included in the message (optional).
    """

    headers: t.Dict[str, str] = field(default_factory=dict)
    data: t.Dict[str, str] = field(default_factory=dict)
    notification: t.Optional[WebpushNotification] = field(default=None)
    fcm_options: t.Optional[WebpushFCMOptions] = field(default=None)


@dataclass
class Message:
    """
    A common message that can be sent via Firebase.

    Contains payload information as well as recipient information. In particular, the message must contain exactly one
    of token, topic or condition fields.

    Attributes:
    data: a dictionary of data fields (optional). All keys and values in the dictionary must be strings.
    notification: an instance of ``messages.Notification`` (optional).
    android: an instance of ``messages.AndroidConfig`` (optional).
    webpush: an instance of ``messages.WebpushConfig`` (optional).
    apns: an instance of ``messages.ApnsConfig`` (optional).
    token: the registration token of the device to which the message should be sent.
    topic: name of the Firebase topic to which the message should be sent (optional).
    condition: the Firebase condition to which the message should be sent (optional).
    """

    token: t.Optional[str]
    data: t.Dict[str, str] = field(default_factory=dict)
    notification: t.Optional[Notification] = field(default=None)
    android: t.Optional[AndroidConfig] = field(default=None)
    webpush: t.Optional[WebpushConfig] = field(default=None)
    apns: t.Optional[APNSConfig] = field(default=None)
    topic: t.Optional[str] = None
    condition: t.Optional[str] = None


@dataclass
class PushNotification:
    """The payload that is sent to Firebase Cloud Messaging.

    Attributes:
    message: an instance of ``messages.Message`` or ``messages.MulticastMessage``.
    validate_only: a boolean indicating whether to run the operation in dry run mode (optional).
    """

    message: Message
    validate_only: t.Optional[bool] = field(default=False)


class FcmPushResponse:
    """The response received from an individual batched request to the FCM API.

    The interface of this object is compatible with SendResponse object of
    the Google's firebase-admin-python package.
    """

    def __init__(
        self, fcm_response: t.Optional[t.Dict[str, str]] = None, exception: t.Optional[AsyncFirebaseError] = None
    ):
        """Inits FcmPushResponse object.

        :param fcm_response: a dictionary with the data that FCM returns as a payload
        :param exception: an exception that may happen when communicating with FCM
        """
        self.message_id = fcm_response.get("name") if fcm_response else None
        self.exception = exception

    @property
    def success(self) -> bool:
        """A boolean indicating if the request was successful."""
        return self.message_id is not None and not self.exception


class FcmPushMulticastResponse:
    """The response received from a batch request to the FCM API.

    The interface of this object is compatible with BatchResponse object of
    the Google's firebase-admin-python package.
    """

    def __init__(self, responses: t.List[FcmPushResponse]):
        """Inits FcmPushMulticastResponse.

        :param responses: a list of FcmPushResponse objects
        """
        self._responses = responses
        self._success_count = len([resp for resp in responses if resp.success])

    @property
    def responses(self):
        """A list of ``FcmPushResponse`` objects (possibly empty)."""
        return self._responses

    @property
    def success_count(self):
        return self._success_count

    @property
    def failure_count(self):
        return len(self.responses) - self.success_count



"""Async Firebase errors."""
import typing as t
from enum import Enum

import httpx


class FcmErrorCode(Enum):
    INVALID_ARGUMENT = "INVALID_ARGUMENT"
    FAILED_PRECONDITION = "FAILED_PRECONDITION"
    OUT_OF_RANGE = "OUT_OF_RANGE"
    UNAUTHENTICATED = "UNAUTHENTICATED"
    PERMISSION_DENIED = "PERMISSION_DENIED"
    NOT_FOUND = "NOT_FOUND"
    CONFLICT = "CONFLICT"
    ABORTED = "ABORTED"
    ALREADY_EXISTS = "ALREADY_EXISTS"
    RESOURCE_EXHAUSTED = "RESOURCE_EXHAUSTED"
    CANCELLED = "CANCELLED"
    DATA_LOSS = "DATA_LOSS"
    UNKNOWN = "UNKNOWN"
    INTERNAL = "INTERNAL"
    UNAVAILABLE = "UNAVAILABLE"
    DEADLINE_EXCEEDED = "DEADLINE_EXCEEDED"


class BaseAsyncFirebaseError(Exception):
    """Base error for Async Firebase"""


class AsyncFirebaseError(BaseAsyncFirebaseError):
    """A prototype for all AF Errors.

    This error and its subtypes and the reason to rise them are consistent with Google's errors,
    that may be found in `firebase-admin-python` in `firebase_admin.exceptions module`.
    """

    def __init__(
        self,
        code: str,
        message: str,
        cause: t.Optional[httpx.HTTPError] = None,
        http_response: t.Optional[httpx.Response] = None,
    ):
        """Init the AsyncFirebase error.

        :param code: A string error code that represents the type of the exception. Possible error
            codes are defined in https://cloud.google.com/apis/design/errors#handling_errors.
        :param message: A human-readable error message string.
        :param cause: The exception that caused this error (optional).
        :param http_response: If this error was caused by an HTTP error response, this property is
            set to the ``httpx.Response`` object that represents the HTTP response (optional).
            See https://www.python-httpx.org/api/#response for details of this object.
        """
        self.code = code
        self.cause = cause
        self.http_response = http_response
        super().__init__(message)


class DeadlineExceededError(AsyncFirebaseError):
    """Request deadline exceeded.

    This will happen only if the caller sets a deadline that is shorter than the method's
    default deadline (i.e. requested deadline is not enough for the server to process the
    request) and the request did not finish within the deadline.
    """

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.DEADLINE_EXCEEDED.value, message, cause=cause, http_response=http_response)


class UnavailableError(AsyncFirebaseError):
    """Service unavailable. Typically the server is down."""

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.UNAVAILABLE.value, message, cause=cause, http_response=http_response)


class UnknownError(AsyncFirebaseError):
    """Unknown server error."""

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.UNKNOWN.value, message, cause=cause, http_response=http_response)


class UnauthenticatedError(AsyncFirebaseError):
    """Request not authenticated due to missing, invalid, or expired OAuth token."""

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.UNAUTHENTICATED.value, message, cause=cause, http_response=http_response)


class ThirdPartyAuthError(UnauthenticatedError):
    """APNs certificate or web push auth key was invalid or missing."""


class ResourceExhaustedError(AsyncFirebaseError):
    """Either out of resource quota or reaching rate limiting."""

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.RESOURCE_EXHAUSTED.value, message, cause=cause, http_response=http_response)


class QuotaExceededError(ResourceExhaustedError):
    """Sending limit exceeded for the message target."""


class PermissionDeniedError(AsyncFirebaseError):
    """Client does not have sufficient permission.

    This can happen because the OAuth token does not have the right scopes, the client doesn't
    have permission, or the API has not been enabled for the client project.
    """

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.PERMISSION_DENIED.value, message, cause=cause, http_response=http_response)


class SenderIdMismatchError(PermissionDeniedError):
    """The authenticated sender ID is different from the sender ID for the registration token."""


class NotFoundError(AsyncFirebaseError):
    """A specified resource is not found, or the request is rejected by undisclosed reasons.

    An example of the possible cause of this error is whitelisting.
    """

    def __init__(self, message, cause=None, http_response=None):
        """Please see params information in the base exception docstring."""
        super().__init__(FcmErrorCode.NOT_FOUND.value, message, cause=cause, http_response=http_response)


class UnregisteredError(NotFoundError):
    """App instance was unregistered from FCM.

    This usually means that the token used is no longer valid and a new one must be used.
    """


class InvalidArgumentError(AsyncFirebaseError):
    """Client specified an invalid argument."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.INVALID_ARGUMENT.value, message, cause=cause, http_response=http_response)


class FailedPreconditionError(AsyncFirebaseError):
    """Request can not be executed in the current system state, such as deleting a non-empty directory."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.FAILED_PRECONDITION.value, message, cause=cause, http_response=http_response)


class OutOfRangeError(AsyncFirebaseError):
    """Client specified an invalid range."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.OUT_OF_RANGE.value, message, cause=cause, http_response=http_response)


class AbortedError(AsyncFirebaseError):
    """Concurrency conflict, such as read-modify-write conflict."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.ABORTED.value, message, cause=cause, http_response=http_response)


class AlreadyExistsError(AsyncFirebaseError):
    """The resource that a client tried to create already exists."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.ALREADY_EXISTS.value, message, cause=cause, http_response=http_response)


class ConflictError(AsyncFirebaseError):
    """Concurrency conflict, such as read-modify-write conflict."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.CONFLICT.value, message, cause=cause, http_response=http_response)


class CancelledError(AsyncFirebaseError):
    """Request cancelled by the client."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.CANCELLED.value, message, cause=cause, http_response=http_response)


class DataLossError(AsyncFirebaseError):
    """Unrecoverable data loss or data corruption."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.DATA_LOSS.value, message, cause=cause, http_response=http_response)


class InternalError(AsyncFirebaseError):
    """Internal server error."""

    def __init__(self, message, cause=None, http_response=None):
        super().__init__(FcmErrorCode.INTERNAL.value, message, cause=cause, http_response=http_response)
